/*
 * Copyright 2008 jpp Development Team
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package net.sourceforge.jpp.handler;

import java.lang.annotation.Annotation;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Set;
import net.sourceforge.jpp.model.TypeName;
import javax.lang.model.element.TypeElement;
import net.sourceforge.jpp.PatternHandler;
import javax.annotation.processing.ProcessingEnvironment;
import javax.lang.model.element.Element;
import javax.lang.model.element.ElementKind;
import javax.lang.model.element.ExecutableElement;
import javax.lang.model.element.PackageElement;
import javax.lang.model.element.TypeParameterElement;
import javax.lang.model.type.DeclaredType;
import javax.lang.model.type.ExecutableType;
import javax.lang.model.type.TypeMirror;
import javax.lang.model.type.TypeVariable;
import javax.tools.Diagnostic;
import net.sourceforge.jpp.annotation.Init;
import net.sourceforge.jpp.model.MethodData;
import net.sourceforge.jpp.model.ModifierData;

public abstract class AbstractPatternHandler<A extends Annotation> implements PatternHandler<A> {

    protected ProcessingEnvironment processingEnv;
    protected Class<A> patternAnnotationType;
    private static final TypeName[] EMPTY_TYPE_NAME_ARRAY = {};
    private static final MethodData[] EMPTY_METHOD_DATA_ARRAY = {};

    public AbstractPatternHandler(Class<A> patternAnnotationType) {
        this.patternAnnotationType = patternAnnotationType;
    }

    @Override
    public Class<A> getPatternAnnotationType() {
        return patternAnnotationType;
    }
    
    @Override
    public Set<String> getSupportedOptions()
    {
        return Collections.EMPTY_SET;
    }

    @Override
    public void init(ProcessingEnvironment processingEnv) {
        this.processingEnv = processingEnv;
    }
    
    protected TypeName newTypeName(String name, TypeElement source, String format) {
        TypeName ret;
        if (name != null && name.length() != 0) {
            ret = new TypeName(name);
        } else {
            PackageElement pkg = processingEnv.getElementUtils().getPackageOf(source);
            ret = new TypeName(String.format(format, pkg == null ? "" : pkg.getQualifiedName().toString(), source.getSimpleName().toString()), getTypeParameters(source));
        }
        return ret;
    }

    protected TypeName newTypeName(TypeElement type) {
        StringBuilder buff = new StringBuilder();
        toSource(type.asType(), buff);
        return new TypeName(buff.toString());
    }

    protected TypeName newTypeName(String name, boolean indexed) {
        return new TypeName(indexed ? name + TypeName.ARRAY_SUFFIX : name);
    }
    
    protected TypeName newTypeName(String name) {
        return new TypeName(name);
    }
    protected TypeName newTypeName(TypeMirror mirror) {
        return new TypeName(mirror.toString());
    }

    protected TypeName[] newTypeNames(List<? extends TypeMirror> values) {
        List<TypeName> converted = new ArrayList<TypeName>();

        for (TypeMirror m : values) {
            converted.add(newTypeName(m));
        }

        return converted.toArray(EMPTY_TYPE_NAME_ARRAY);
    }

    protected String getTypeParameters(TypeElement type) {
        DeclaredType dt = (DeclaredType) type.asType();

        StringBuilder buff = new StringBuilder();
        List<? extends TypeMirror> args = dt.getTypeArguments();

        if (!args.isEmpty()) {
            buff.append('<');


            for (TypeMirror var : args) {
                toSource((TypeVariable) var, buff);
                buff.append(',');
            }
            buff.deleteCharAt(buff.length() - 1);
            buff.append('>');
        }

        return buff.toString();
    }

    protected String getTypeParameters(ExecutableElement element) {
        ExecutableType et = (ExecutableType) element.asType();
        StringBuilder buff = new StringBuilder();

        List<? extends TypeMirror> args = et.getTypeVariables();

        if (!args.isEmpty()) {
            buff.append('<');

            for (TypeMirror v : et.getTypeVariables()) {
                toSource((TypeVariable) v, buff);
                buff.append(',');
            }
            buff.deleteCharAt(buff.length() - 1);
            buff.append('>');
        }

        return buff.toString();
    }

    protected void printMessage(Diagnostic.Kind kind, CharSequence message) {
        processingEnv.getMessager().printMessage(kind, message);
    }

    protected String getTypeParameters(TypeMirror mirror) {
        StringBuilder buff = new StringBuilder();
        toSource(mirror, buff);
        return buff.toString();
    }

    protected void toSource(TypeMirror mirror, StringBuilder buff) {
        switch (mirror.getKind()) {
            case TYPEVAR:
                toSource((TypeVariable) mirror, buff);
                break;

            case DECLARED:
                toSource((DeclaredType) mirror, buff);
                break;

            default:
                buff.append(mirror.toString());
        }
    }

    protected void toSource(DeclaredType mirror, StringBuilder buff) {
        buff.append(((TypeElement) mirror.asElement()).getQualifiedName());

        List<? extends TypeMirror> args = mirror.getTypeArguments();

        if (!args.isEmpty()) {
            buff.append('<');
            for (TypeMirror var : args) {
                TypeVariable v = (TypeVariable) var;
                buff.append(v.toString());
                buff.append(',');
            }
            buff.deleteCharAt(buff.length() - 1);
            buff.append('>');
        }
    }

    protected void toSource(TypeVariable mirror, StringBuilder buff) {
        buff.append(mirror.toString());

        TypeParameterElement element = (TypeParameterElement) mirror.asElement();
        List<? extends TypeMirror> bounds = element.getBounds();

        if (!bounds.isEmpty() && !(bounds.size() == 1 && Object.class.getName().equals(bounds.get(0).toString()))) {

            buff.append(" extends ");
            for (TypeMirror m : bounds) {
                toSource(m, buff);
                buff.append(',');
            }
            buff.deleteCharAt(buff.length() - 1);
        }
    }

    protected MethodData[] listMethods(TypeElement type) {
        List<MethodData> methods = new ArrayList<MethodData>();
        for (Element e : type.getEnclosedElements()) {
            if (e.getKind() != ElementKind.METHOD) {
                continue;
            }
            methods.add(getMethodData((ExecutableElement) e));
        }
        return methods.toArray(EMPTY_METHOD_DATA_ARRAY);
    }

    protected MethodData getMethodData(ExecutableElement element) {
        MethodData method = new MethodData();
        method.setName(element.getSimpleName().toString());

        ExecutableType t = (ExecutableType) element.asType();
        method.setParameterTypes(newTypeNames(t.getParameterTypes()));
        method.setExceptionTypes(newTypeNames(t.getThrownTypes()));

        method.setReturnType(newTypeName(t.getReturnType()));

        method.setTypeParameters(getTypeParameters(element));
        
        if ( ! element.getModifiers().isEmpty() )
        {
            method.setModifiers(new ModifierData(element.getModifiers()));
        }

        return method;
    }

    protected MethodData getInit(TypeElement type)
    {
        MethodData init = null;
        for(Element e : processingEnv.getElementUtils().getAllMembers(type))
        {
            if ( e.getAnnotation(Init.class) != null )
            {
                init = getMethodData((ExecutableElement) e);
                break;
            }
        }
        return init;
    }   
}
