/*
 * Copyright 2008 jpp Development Team
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package net.sourceforge.jpp.handler;

import java.beans.PropertyChangeListener;
import java.beans.PropertyVetoException;
import java.beans.VetoableChangeListener;
import javax.lang.model.element.Modifier;
import javax.lang.model.element.TypeElement;
import net.sourceforge.jpp.annotation.EventSet;
import net.sourceforge.jpp.annotation.Bean;
import net.sourceforge.jpp.annotation.Field;
import net.sourceforge.jpp.annotation.Method;
import net.sourceforge.jpp.annotation.Property;
import net.sourceforge.jpp.model.EventSetData;
import net.sourceforge.jpp.model.BeanData;
import net.sourceforge.jpp.model.FieldData;
import net.sourceforge.jpp.model.MethodData;
import net.sourceforge.jpp.model.ModifierData;
import net.sourceforge.jpp.model.PropertyData;
import net.sourceforge.jpp.model.TypeName;

public class BeanHandler extends SourceGeneratorPatternHandler<Bean> {

    public static final String EVENT_OBJECT_SUFFIX = "Event";
    public static final String EVENT_LISTENER_SUFFIX = "Listener";
    public static final String PROP_WRITE_PREFIX = "set";
    public static final String PROP_READ_PREFIX = "get";
    public static final String PROP_READ_BOOLEAN_PREFIX = "is";
    public static final String ADD_LISTENER_PREFIX = "add";
    public static final String REMOVE_LISTENER_PREFIX = "remove";
    public static final String FIRE_EVENT_PREFIX = "fire";
    public static final String PROP_NAME_CONSTANT_PREFIX = "PROP_";
    public static final char PROP_NAME_CONSTANT_SEPARATOR = '_';
    public static final String PROP_FIELD_NAME_PREFIX = "_";
    
    private static final TypeName[] _constrainedException = { new TypeName("java.beans.PropertyVetoException")};
    private static final TypeName[] _unicastException = { new TypeName("java.util.TooManyListenersException")};
    private static final TypeName[] _exceptionHandlerParameters = { new TypeName("java.lang.Exception")};

    public BeanHandler() {
        super(Bean.class);
    }

    @Override
    public BeanData buildModel(TypeElement type) {

        Bean annotation = type.getAnnotation(Bean.class);
        BeanData model = new BeanData();

        model.getTarget().setName(newTypeName(annotation.name()));
        model.getSource().setName(newTypeName(type));
        model.setConcrete(annotation.concrete());
        model.setThreaded(annotation.threaded());
        
        model.setListenerList(buildFieldData(annotation.listenerList()));
        model.setPropertyChangeSupport(buildFieldData(annotation.propertyChangeSupport()));
        model.setVetoableChangeSupport(buildFieldData(annotation.vetoableChangeSupport()));
        
        model.setEventSets(buildEventSets(annotation.eventsets()));
        model.setProperties(buildProperties(annotation.properties()));
        
        model.setExceptionHandler(buildExceptionHandler(annotation.exceptionHandler()));
        
        model.setAddPropertyChangeListenerMethod(buildChangeListenerMethod(PropertyChangeListener.class, annotation.addPropertyChangeListener()));
        model.setRemovePropertyChangeListenerMethod(buildChangeListenerMethod(PropertyChangeListener.class, annotation.removePropertyChangeListener()));
        model.setAddVetoableChangeListenerMethod(buildChangeListenerMethod(VetoableChangeListener.class, annotation.addVetoableChangeListener()));
        model.setRemoveVetoableChangeListenerMethod(buildChangeListenerMethod(VetoableChangeListener.class, annotation.removeVetobaleChangeListener()));
        
        return model;
    }

    protected MethodData buildExceptionHandler(Method annotation)
    {
        MethodData method = new MethodData();
        method.setName(annotation.name());
        method.setInherited(annotation.inherited());
        method.setModifiers(new ModifierData(Modifier.PROTECTED));
        method.setParameterTypes(_exceptionHandlerParameters);
        method.setReturnType(newTypeName(void.class.getName()));
        return method;
    }
    
    protected FieldData buildFieldData(Field annotation)
    {
        FieldData field = new FieldData();
        field.setName(annotation.name());
        field.setInherited(annotation.inherited());
        field.setType(newTypeName(annotation.type()));
        field.setImpl(annotation.impl().length() == 0 ? field.getType() : newTypeName(annotation.impl()) );
        field.setModifiers(new ModifierData(Modifier.PROTECTED));
        return field;
    }
    
    protected EventSetData[] buildEventSets(EventSet[] annotations) {
        EventSetData[] eventSets = new EventSetData[annotations.length];
        for (int i = 0; i < eventSets.length; i++) {
            eventSets[i] = buildEventSet(annotations[i]);
        }
        return eventSets;
    }

    protected EventSetData buildEventSet(EventSet annotation) {
        EventSetData eventSet = new EventSetData();
        eventSet.setName(annotation.name());
        eventSet.setUnicast(annotation.unicast());
        eventSet.setListener(newTypeName(annotation.listener()));
        eventSet.setAddListenerMethod(buildAddListenerMethod(eventSet));
        eventSet.setRemoveListenerMethod(buildRemoveListenerMethod(eventSet));
        
        TypeElement listener = processingEnv.getElementUtils().getTypeElement(eventSet.getListener().getQualifiedName());
        
        MethodData[] listenerMethods = listMethods(listener);
        
        MethodData[] fireEventMethods = new MethodData[listenerMethods.length];
        
        TypeElement event;
        
        if ( annotation.event().length() != 0 )
        {
            event = processingEnv.getElementUtils().getTypeElement(annotation.event());
        }
        else if ( listenerMethods.length > 0 && listenerMethods[0].getParameterTypes().length > 0)
        {
            event = processingEnv.getElementUtils().getTypeElement(listenerMethods[0].getParameterTypes()[0].getQualifiedName());
        }
        else
        {
            event = null;
        }
                
        if ( event != null )
        {
            eventSet.setEvent(newTypeName(event));
        }
        
        for(int i = 0; i < listenerMethods.length; i++)
        {
            fireEventMethods[i] = buildFireEventMethod(eventSet, listenerMethods[i]);
        }
        
        eventSet.setlistenerMethods(listenerMethods);
        eventSet.setFireEventMethods(fireEventMethods);

        return eventSet;
    }

    protected MethodData buildFireEventMethod(EventSetData eventSet, MethodData listenerMethod)
    {
        MethodData method = new MethodData();
        method.setName(fireMethod(eventSet.getName(), listenerMethod.getName()));
        method.setReturnType(newTypeName(void.class.getName()));
        method.setParameterTypes(new TypeName[] { eventSet.getEvent()});
        method.setModifiers(new ModifierData(Modifier.PROTECTED));
        return method;
    }
    
    protected PropertyData[] buildProperties(Property[] annotations) {
        PropertyData[] properties = new PropertyData[annotations.length];
        for (int i = 0; i < properties.length; i++) {
            properties[i] = buildProperty(annotations[i]);
        }
        return properties;
    }

    protected PropertyData buildProperty(Property annotation) {
        PropertyData property = new PropertyData();
        property.setName(annotation.name());
        property.setBound(annotation.bound());
        property.setIndexed(annotation.indexed());
        property.setConstrained(annotation.constrained());
        property.setType(newTypeName(annotation.type(),annotation.indexed()));
        property.setImpl(annotation.impl().length() == 0 ? property.getType() : newTypeName(annotation.impl(),annotation.indexed()));
        property.setWriteMethod(buildWriteMethod(property,annotation));
        property.setReadMethod(buildReadMethod(property, annotation));
        
        StringBuilder buff = new StringBuilder(PROP_NAME_CONSTANT_PREFIX);
        for(char c: property.getName().toCharArray())
        {
            if ( Character.isUpperCase(c) )
            {
                buff.append(PROP_NAME_CONSTANT_SEPARATOR);
                buff.append(c);
            }
            else
            {
                buff.append(Character.toUpperCase(c));
            }
        }
        property.setNameConstant(buff.toString());
        
        FieldData field = new FieldData();
        field.setInherited(false);
        field.setModifiers(new ModifierData(Modifier.PRIVATE));
        field.setName(PROP_FIELD_NAME_PREFIX + property.getName());
        field.setType(property.getType());
        field.setImpl(property.getImpl());
        property.setField(field);
        
        return property;
    }
    
    private String fireMethod(String eventSetName, String methodName)
    {
        StringBuilder buff = new StringBuilder();
        buff.append(FIRE_EVENT_PREFIX);
        buff.append(Character.toUpperCase(eventSetName.charAt(0)));
        buff.append(eventSetName.substring(1));
        buff.append(Character.toUpperCase(methodName.charAt(0)));
        buff.append(methodName.substring(1));
        return buff.toString();
    }
    
    private String propertyMethod(String prefix, String property)
    {
        StringBuilder buff = new StringBuilder();
        buff.append(prefix);
        buff.append(Character.toUpperCase(property.charAt(0)));
        buff.append(property.substring(1));
        return buff.toString();
    }
    
    private String listenerMethod(String prefix, TypeName listener)
    {
        StringBuilder buff = new StringBuilder();
        buff.append(prefix);
        buff.append(listener.getSimpleBaseName());
        return buff.toString();
    }
    
    protected MethodData buildReadMethod(PropertyData property, Property annotation)
    {
        MethodData method = new MethodData();
        method.setName( propertyMethod(boolean.class.getName().equals(property.getType().getSimpleName()) ? PROP_READ_BOOLEAN_PREFIX : PROP_READ_PREFIX ,property.getName()));
        method.setReturnType(property.getType());
        method.setParameterTypes(new TypeName[] {});
        if (annotation.read().hasModifier())
        {
            method.setModifiers(new ModifierData(annotation.read().getModifier() ));
        }
        return method;
    }

    protected MethodData buildWriteMethod(PropertyData property, Property annotation)
    {
        MethodData method = new MethodData();
        method.setName( propertyMethod(PROP_WRITE_PREFIX,property.getName()));
        method.setReturnType(newTypeName(void.class.getName()));
        method.setParameterTypes(new TypeName[] { property.getType() });
        if ( property.isConstrained() )
        {
            method.setExceptionTypes(new TypeName[] {newTypeName(PropertyVetoException.class.getName())});
        }
        if (annotation.write().hasModifier())
        {
            method.setModifiers(new ModifierData(annotation.write().getModifier() ));
        }
        if ( property.isConstrained())
        {
            method.setExceptionTypes(_constrainedException);
        }
        return method;
    }
    
    protected MethodData buildAddListenerMethod(EventSetData eventSet)
    {
        MethodData method = new MethodData();
        method.setName( listenerMethod(ADD_LISTENER_PREFIX,eventSet.getListener()));
        method.setReturnType(newTypeName(void.class.getName()));
        method.setParameterTypes(new TypeName[] { eventSet.getListener() });
        if ( eventSet.isUnicast() )
        {
            method.setExceptionTypes(_unicastException);
        }
        method.setModifiers(new ModifierData(Modifier.PUBLIC));
        return method;
    }
    
    protected MethodData buildRemoveListenerMethod(EventSetData eventSet)
    {
        MethodData method = new MethodData();
        method.setName( listenerMethod(REMOVE_LISTENER_PREFIX,eventSet.getListener()));
        method.setReturnType(newTypeName(void.class.getName()));
        method.setParameterTypes(new TypeName[] { eventSet.getListener() });
        method.setModifiers(new ModifierData(Modifier.PUBLIC));
        return method;
    }
    
    protected MethodData buildChangeListenerMethod(Class listener, Method annotation)
    {
        MethodData method = new MethodData();
        method.setName(annotation.name());
        method.setInherited(annotation.inherited());
        method.setReturnType(newTypeName(void.class.getName()));
        method.setParameterTypes(new TypeName[] { newTypeName(String.class.getName()), newTypeName(listener.getName()) });
        method.setModifiers(new ModifierData(Modifier.PUBLIC));
        return method;
    }
}
