/*
 * Copyright 2008 jpp Development Team
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package net.sourceforge.jpp.handler;

import freemarker.cache.ClassTemplateLoader;
import freemarker.template.Configuration;
import freemarker.template.DefaultObjectWrapper;
import freemarker.template.Template;
import freemarker.template.TemplateException;
import java.io.IOException;
import java.io.Writer;
import java.lang.annotation.Annotation;
import java.util.HashSet;
import java.util.Set;
import javax.annotation.processing.ProcessingEnvironment;
import javax.lang.model.element.TypeElement;
import javax.tools.Diagnostic.Kind;
import javax.tools.JavaFileObject;
import net.sourceforge.jpp.PatternHandler;
import net.sourceforge.jpp.model.SourceGeneratorData;
import net.sourceforge.jpp.model.TemplateRoot;

public abstract class SourceGeneratorPatternHandler<A extends Annotation> extends AbstractPatternHandler<A> {

    public static final String PACKAGE_ROOT_PATH = "/";
    public static final String DEFAULT_TEMPLATE_PACKAGE = "template";
    public static final String DEFAULT_TEMPLATE_SUFFIX = ".ftl";
    public static final String OPT_TEMPLATE_SUFFIX = ".template";
    private Template _template;

    public SourceGeneratorPatternHandler(Class<A> patternAnnotationType) {
        super(patternAnnotationType);
    }

    @Override
    public void handlePattern(TypeElement type) {

        generateSource(buildModel(type), type);
    }

    protected abstract SourceGeneratorData buildModel(TypeElement type);

    protected void generateSource(SourceGeneratorData model, TypeElement type) {

        printMessage(Kind.NOTE, "Generating source for " + model.getTarget().getName().getQualifiedName());

        try {

            JavaFileObject file = processingEnv.getFiler().createSourceFile(model.getTarget().getName().getQualifiedBaseName(), type);

            Writer out = file.openWriter();

            _template.process(new TemplateRoot(model), out);

            out.close();
        } catch (IOException x) {
            printMessage(Kind.ERROR, x.toString());
        } catch (TemplateException x) {
            printMessage(Kind.ERROR, x.toString());
        }
    }

    @Override public Set<String> getSupportedOptions()
    {
        Set<String> ret = new HashSet<String>();
        ret.add(this.patternAnnotationType.getName() + OPT_TEMPLATE_SUFFIX);
        ret.add(this.patternAnnotationType.getSimpleName() + OPT_TEMPLATE_SUFFIX);
        return ret;
    }
    @Override
    public void init(ProcessingEnvironment processingEnv) {

        super.init(processingEnv);
        Configuration cfg = new Configuration();
        cfg.setObjectWrapper(new DefaultObjectWrapper());
        String templateName = this.processingEnv.getOptions().get(this.patternAnnotationType.getName() + OPT_TEMPLATE_SUFFIX);
        if ( templateName == null )
        {
            templateName = this.processingEnv.getOptions().get(this.patternAnnotationType.getSimpleName() + OPT_TEMPLATE_SUFFIX);
        }        
        cfg.setTemplateLoader(new ClassTemplateLoader(PatternHandler.class, templateName == null ? DEFAULT_TEMPLATE_PACKAGE : PACKAGE_ROOT_PATH));
        if ( templateName == null )
        {
            templateName = this.patternAnnotationType.getSimpleName() + DEFAULT_TEMPLATE_SUFFIX ;
        }
        try {
            _template = cfg.getTemplate(templateName);
        } catch (IOException x) {
            printMessage(Kind.ERROR, x.toString());
        }
        _template.setEncoding("UTF8");
    }
}
